import logging
import aiofiles
import os

appliances_dict = {}
mqtt_appliances_dict = {}
ack_events = {}
loggers = {}

def get_appliance(device_id: tuple, channel: tuple, appliances_dict: dict):
    # appliance key form is ((device_id),(channels))
    try:
        device_appliances = [
            a
            for k, a in appliances_dict.items()
            if k[0] == device_id and channel in k[1]
        ]
        return tuple(device_appliances)[0]
    except IndexError:
        logging.error(f"No appliances found for key {(device_id,channel)}")
        return None
    except Exception as e:
        logging.error(f"An error occurred: {e}")
        return None

async def get_real_mac(interface='end0') -> str | None:
    path = f'/sys/class/net/{interface}/address'
    if not os.path.exists(path):
        return None
    try:
        async with aiofiles.open(path, mode='r') as f:
            mac = await f.read()
            return mac.strip().upper()
    except Exception:
        return None



from homeassistant.core import HomeAssistant

import logging
import asyncio
from TISControlProtocol.shared import ack_events


async def handle_real_time_feedback(hass: HomeAssistant, info: dict):
    channel_number = info["additional_bytes"][0]
    # await target_appliance.handle_packet(info["additional_bytes"], "control")
    if info["source_device_id"] == [0x64, 0x64]:
        event_data = {
            "device_id": info["device_id"],
            "channel_number": channel_number,
            "feedback_type": "realtime_feedback",
            "additional_bytes": info["additional_bytes"],
        }
        try:
            hass.bus.async_fire(str(info["device_id"]), event_data)
        except Exception as e:
            logging.error(f"error in firing event for feedback: {e}")

    # try:
    #     event: asyncio.Event = ack_events.get(
    #         (
    #             tuple(info["device_id"]),
    #             (0x00, 0x31),
    #             channel_number,
    #         )
    #     )
    #     if event is not None:
    #         print("setting event")
    #         event.set()
    # except Exception as e:
    #     print(e)



from TISControlProtocol.Protocols import setup_udp_protocol
from TISControlProtocol.Protocols.udp.ProtocolHandler import (
    TISProtocolHandler,
    TISPacket,
)

import os
from datetime import timedelta
from homeassistant.helpers.event import async_track_time_interval
from homeassistant.core import HomeAssistant
from homeassistant.components.http import HomeAssistantView
from homeassistant.helpers.aiohttp_client import async_get_clientsession
from typing import Optional
import aiohttp
from aiohttp import web
import aiofiles
import socket
import logging
from collections import defaultdict
import json
import psutil
import asyncio
import ST7789
from PIL import Image, ImageDraw, ImageFont
from TISControlProtocol.shared import get_real_mac

protocol_handler = TISProtocolHandler()


class TISApi:
    """TIS API class."""

    def __init__(
        self,
        port: int,
        hass: HomeAssistant,
        domain: str,
        devices_dict: dict,
        version: str,
        host: str = "0.0.0.0",
        display_logo: Optional[str] = None,
    ):
        """Initialize the API class."""
        self.host = host
        self.port = port
        self.protocol = None
        self.transport = None
        self.hass = hass
        self.config_entries = {}
        self.bill_configs = {}
        self.domain = domain
        self.devices_dict = devices_dict
        self.display_logo = display_logo
        self.display = None
        self.version = version
        self.cms_url = "https://cms-tis.com"

    async def connect(self):
        """Connect to the TIS API."""
        self.loop = self.hass.loop
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
            await self._setup_udp_protocol()
            await self._initialize_hass_data()
            await self._register_http_views()
            self.hass.async_add_executor_job(self.run_display)
            self._register_services()
            self._schedule_cms_data_task()
        except Exception as e:
            logging.error("Error during connection setup: %s", e)
            raise ConnectionError

    async def _setup_udp_protocol(self):
        """Setup the UDP protocol."""
        try:
            self.transport, self.protocol = await setup_udp_protocol(
                self.sock,
                self.loop,
                self.host,
                self.port,
                self.hass,
            )
        except Exception as e:
            logging.error("Error connecting to TIS API %s", e)
            raise ConnectionError

    async def _initialize_hass_data(self):
        """Initialize Home Assistant data."""
        self.hass.data[self.domain]["discovered_devices"] = []

    async def _register_http_views(self):
        """Register HTTP views."""
        try:
            self.hass.http.register_view(TISEndPoint(self))
            self.hass.http.register_view(ScanDevicesEndPoint(self))
            self.hass.http.register_view(GetKeyEndpoint(self))
            self.hass.http.register_view(ChangeSecurityPassEndpoint(self))
            self.hass.http.register_view(RestartEndpoint(self))
            self.hass.http.register_view(UpdateEndpoint(self))
            self.hass.http.register_view(BillConfigEndpoint(self))
            self.hass.http.register_view(GetBillConfigEndpoint(self))
        except Exception as e:
            logging.error("Error registering views %s", e)
            raise ConnectionError

    def _register_services(self):
        """Register Home Assistant services."""

        async def handle_cms_data(call):
            data = call.data.get("data", None)

            if data is None:
                logging.error("No data provided to send to CMS")
                return

            cms_sender = CMSDataSender(
                external_url=f"{self.cms_url}/api/device-health",
                hass=self.hass,
            )
            await cms_sender.send_data(data)

        self.hass.services.async_register(
            self.domain,
            "send_cms_data",
            handle_cms_data,
        )

    def _schedule_cms_data_task(self):
        """Schedule periodic CMS data task."""

        async def scheduled_task(now=None):
            try:
                data = await self._collect_system_data()

                await self.hass.services.async_call(
                    self.domain,
                    "send_cms_data",
                    {"data": data},
                )
            except Exception as e:
                logging.error(f"Error getting data for CMS: {e}")

        interval = timedelta(minutes=3)
        async_track_time_interval(self.hass, scheduled_task, interval)

    async def _collect_system_data(self):
        """Collect system data for CMS."""
        # Mac Address
        mac_address = await get_real_mac("end0")

        # CPU Usage
        cpu_usage = await self.hass.async_add_executor_job(psutil.cpu_percent, 1)

        # CPU Temperature
        cpu_temp = await self.hass.async_add_executor_job(psutil.sensors_temperatures)
        cpu_temp = cpu_temp.get("cpu_thermal", None)
        cpu_temp = cpu_temp[0].current if cpu_temp else 0

        # Disk Usage
        total, _, free, percent = await self.hass.async_add_executor_job(
            psutil.disk_usage, "/"
        )

        # Memory Usage
        mem = await self.hass.async_add_executor_job(psutil.virtual_memory)

        return {
            "mac_address": mac_address,
            "cpu_usage": cpu_usage,
            "cpu_temperature": cpu_temp,
            "disk_total": total,
            "disk_free": free,
            "disk_percent": percent,
            "ram_total": mem.total,
            "ram_free": mem.free,
            "ram_percent": mem.percent,
        }

    def run_display(self, style="dots"):
        try:
            self.display = ST7789.ST7789(
                width=320,
                height=240,
                rotation=0,
                port=0,
                cs=0,
                dc=23,
                rst=25,
                backlight=12,
                spi_speed_hz=60 * 1000 * 1000,
                offset_left=0,
                offset_top=0,
            )
            # Initialize display.
            self.display.begin()
            self.set_display_image()

        except Exception as e:
            logging.error(f"error initializing display, {e}")
            return

    def set_display_image(self):
        if self.display_logo:
            img = Image.open(self.display_logo).convert("RGB")
            version_text = f"V {self.version}"

            draw = ImageDraw.Draw(img)
            font = ImageFont.load_default(size=28)
            x, y = 78, 235
            draw.text((x, y), version_text, font=font, fill=(255, 255, 255))
            img = img.rotate(-90, expand=True)

            self.display.set_backlight(0)
            self.display.display(img)

    async def parse_device_manager_request(self, data: dict) -> None:
        """Parse the device manager request."""
        converted = {
            appliance: {
                "device_id": [int(n) for n in details[0]["device_id"].split(",")],
                "appliance_type": details[0]["appliance_type"]
                .lower()
                .replace(" ", "_"),
                "appliance_class": details[0].get("appliance_class", None),
                "is_protected": bool(int(details[0]["is_protected"])),
                "gateway": details[0]["gateway"],
                "channels": [
                    {
                        "channel_number": int(detail["channel_number"]),
                        "channel_name": detail["channel_name"],
                    }
                    for detail in details
                ],
                "min": details[0]["min"],
                "max": details[0]["max"],
                "settings": details[0]["settings"],
            }
            for appliance, details in data["appliances"].items()
        }

        grouped = defaultdict(list)
        for appliance, details in converted.items():
            grouped[details["appliance_type"]].append({appliance: details})
        self.config_entries = dict(grouped)

        # add a lock module config entry
        self.config_entries["lock_module"] = {
            "password": data["configs"]["lock_module_password"]
        }
        return self.config_entries

    async def get_entities(self, platform: str = None) -> list:
        """Get the stored entities."""
        directory = "/config/custom_components/tis_integration/"
        os.makedirs(directory, exist_ok=True)

        data = await self.read_appliances(directory)

        await self.parse_device_manager_request(data)
        entities = self.config_entries.get(platform, [])
        return entities

    async def read_appliances(self, directory: str) -> dict:
        """Read, decrypt, and return the stored data."""
        file_name = "app.json"
        output_file = os.path.join(directory, file_name)

        try:
            async with aiofiles.open(output_file, "r") as f:
                raw_data = await f.read()
                # logging.warning(f"file length: {len(raw_data)}")
                if raw_data:
                    encrypted_data = json.loads(raw_data)
                    data = self.decrypt_data(encrypted_data)
                else:
                    data = {}
        except FileNotFoundError:
            data = {}
        return data

    async def save_appliances(self, data: dict, directory: str) -> None:
        """Encrypt and save the data."""
        file_name = "app.json"
        output_file = os.path.join(directory, file_name)

        encrypted_data = self.encrypt_data(data)
        logging.warning(f"file (to be saved) length: {len(encrypted_data)}")

        async with aiofiles.open(output_file, "w") as f:
            logging.warning("new appliances are getting saved in app.json")
            await f.write(json.dumps(encrypted_data, indent=4))

        logging.warning("new applinaces saved successfully")

    async def get_bill_configs(self) -> dict:
        """Get Bill Configurations"""
        try:
            directory = "/config/custom_components/tis_integration/"
            os.makedirs(directory, exist_ok=True)

            file_name = "bill.json"
            output_file = os.path.join(directory, file_name)

            async with aiofiles.open(output_file, "r") as f:
                data = json.loads(await f.read())
        except FileNotFoundError:
            async with aiofiles.open(output_file, "w") as f:
                await f.write(json.dumps(""))
                data = {}
        self.bill_configs = data
        return data

    def encrypt(self, text: str, shift: int = 5) -> str:
        result = ""
        for char in text:
            if char.isalpha():
                base = ord("A") if char.isupper() else ord("a")
                result += chr((ord(char) - base + shift) % 26 + base)
            else:
                result += char
        return result

    def decrypt(self, text: str, shift: int = 5) -> str:
        return self.encrypt(text, -shift)

    def encrypt_data(self, data, shift: int = 5):
        if isinstance(data, dict):
            return {
                self.encrypt(str(k), shift): self.encrypt_data(v, shift)
                for k, v in data.items()
            }
        elif isinstance(data, list):
            return [self.encrypt_data(item, shift) for item in data]
        elif isinstance(data, str):
            return self.encrypt(data, shift)
        else:
            return data

    def decrypt_data(self, data, shift: int = 5):
        if isinstance(data, dict):
            return {
                self.decrypt(str(k), shift): self.decrypt_data(v, shift)
                for k, v in data.items()
            }
        elif isinstance(data, list):
            return [self.decrypt_data(item, shift) for item in data]
        elif isinstance(data, str):
            return self.decrypt(data, shift)
        else:
            return data


class TISEndPoint(HomeAssistantView):
    """TIS API endpoint."""

    url = "/api/tis"
    name = "api:tis"
    requires_auth = False

    def __init__(self, tis_api: TISApi):
        """Initialize the API endpoint."""
        self.api = tis_api

    async def post(self, request):
        directory = "/config/custom_components/tis_integration/"

        # Parse the JSON data from the request
        data = await request.json()
        await self.api.save_appliances(data, directory)

        # Start reload operations in the background
        asyncio.create_task(self.reload_platforms())

        # Return the response immediately
        return web.json_response({"message": "success"})

    async def reload_platforms(self):
        # Reload the platforms
        for entry in self.api.hass.config_entries.async_entries(self.api.domain):
            await self.api.hass.config_entries.async_reload(entry.entry_id)


class ScanDevicesEndPoint(HomeAssistantView):
    """Scan Devices API endpoint."""

    url = "/api/scan_devices"
    name = "api:scan_devices"
    requires_auth = False

    def __init__(self, tis_api: TISApi):
        """Initialize the API endpoint."""
        self.api = tis_api
        self.discovery_packet: TISPacket = protocol_handler.generate_discovery_packet()

    async def get(self, request):
        # Discover network devices
        devices = await self.discover_network_devices()
        devices = [
            {
                "device_id": device["device_id"],
                "device_type_code": device["device_type"],
                "device_type_name": self.api.devices_dict.get(
                    tuple(device["device_type"]), tuple(device["device_type"])
                ),
                "gateway": device["source_ip"],
            }
            for device in devices
        ]
        return web.json_response(devices)

    async def discover_network_devices(self, broadcast_attempts=30) -> list:
        # empty current discovered devices list
        self.api.hass.data[self.api.domain]["discovered_devices"] = []
        for i in range(broadcast_attempts):
            await self.api.protocol.sender.broadcast_packet(self.discovery_packet)
            # sleep for 1 sec
            await asyncio.sleep(1)

        return self.api.hass.data[self.api.domain]["discovered_devices"]


class GetKeyEndpoint(HomeAssistantView):
    """Get Key API endpoint."""

    url = "/api/get_key"
    name = "api:get_key"
    requires_auth = False

    def __init__(self, tis_api: TISApi):
        """Initialize the API endpoint."""
        self.api = tis_api

    async def get(self, request):
        # Get the MAC address
        mac_address = await get_real_mac("end0")
        if mac_address is None:
            return web.json_response(
                {"error": "Could not retrieve MAC address"}, status=500
            )

        # Return the MAC address
        return web.json_response({"key": mac_address})


class ChangeSecurityPassEndpoint(HomeAssistantView):
    """Change Security Password API Endpoint."""

    url = "/api/change_pass"
    name = "api:change_pass"
    requires_auth = False

    def __init__(self, tis_api: TISApi):
        self.tis_api = tis_api

    async def post(self, request):
        try:
            old_pass = request.query.get("old_pass")
            new_pass = request.query.get("new_pass")
            confirm_pass = request.query.get("confirm_pass")

            if old_pass is None or new_pass is None or confirm_pass is None:
                logging.info(
                    "Required parameters not found in query, parsing request body"
                )
                data = await request.json()
                old_pass = old_pass or data.get("old_pass")
                new_pass = new_pass or data.get("new_pass")
                confirm_pass = confirm_pass or data.get("confirm_pass")

            if old_pass is None or new_pass is None or confirm_pass is None:
                logging.error("Missing required parameters")
                return web.json_response(
                    {
                        "message": "error",
                        "error": "Missing required parameters",
                    },
                    status=400,
                )

        except Exception as e:
            logging.error(f"Error parsing request: {e}")
            return web.json_response(
                {"message": "error", "error": "Invalid request parameters"},
                status=400,
            )

        if old_pass != self.tis_api.config_entries["lock_module"]["password"]:
            return web.json_response(
                {
                    "message": "error",
                    "error": "Old password is incorrect, please try again",
                },
                status=403,
            )

        if new_pass == old_pass:
            return web.json_response(
                {
                    "message": "error",
                    "error": "New password must be different from the old password",
                },
                status=400,
            )

        if len(new_pass) < 4:
            return web.json_response(
                {
                    "message": "error",
                    "error": "Password must be at least 4 characters long",
                },
                status=400,
            )

        if new_pass != confirm_pass:
            return web.json_response(
                {
                    "message": "error",
                    "error": "New password and confirmation do not match",
                },
                status=400,
            )

        directory = "/config/custom_components/tis_integration/"
        data = await self.tis_api.read_appliances(directory=directory)
        data["configs"]["lock_module_password"] = new_pass
        await self.tis_api.save_appliances(data, directory)
        self.tis_api.config_entries["lock_module"]["password"] = new_pass

        asyncio.create_task(self.reload_platforms())

        return web.json_response(
            {
                "message": "success",
            }
        )

    async def reload_platforms(self):
        for entry in self.tis_api.hass.config_entries.async_entries(
            self.tis_api.domain
        ):
            await self.tis_api.hass.config_entries.async_reload(entry.entry_id)


class RestartEndpoint(HomeAssistantView):
    """Restart the Server"""

    url = "/api/restart"
    name = "api:restart"
    requires_auth = False

    def __init__(self, tis_api: TISApi):
        self.tis_api = tis_api

    async def post(self, request):
        mac_address = request.query.get("mac_address")

        if mac_address is None:
            logging.info("Required parameters not found in query, parsing request body")
            data = await request.json()
            mac_address = data.get("mac_address")

        mac = await get_real_mac("end0")

        if mac_address is None:
            return web.json_response(
                {"error": "required parameters are missing"}, status=400
            )
        elif mac_address != mac:
            return web.json_response({"error": "Unauthorized"}, status=403)

        logging.info("Restarting Server")
        try:
            await self.tis_api.hass.services.async_call(
                "homeassistant", "restart", {}, blocking=False
            )
            return web.json_response({"message": "Server is restarting"}, status=200)
        except Exception as e:
            logging.error(f"Error restarting server: {e}")
            return web.json_response({"error": "Failed to restart server"}, status=500)


class UpdateEndpoint(HomeAssistantView):
    """Update the Server"""

    url = "/api/update"
    name = "api:update"
    requires_auth = False

    def __init__(self, tis_api: TISApi):
        self.tis_api = tis_api

    async def post(self, request):
        mac_address = request.query.get("mac_address")

        if mac_address is None:
            logging.info("Required parameters not found in query, parsing request body")
            data = await request.json()
            mac_address = data.get("mac_address")

        mac = await get_real_mac("end0")

        if mac_address is None:
            return web.json_response(
                {"error": "required parameters are missing"}, status=400
            )
        elif mac_address != mac:
            return web.json_response({"error": "Unauthorized"}, status=403)

        try:
            integration_dir = self.tis_api.hass.config.path(
                "custom_components", "tis_integration"
            )

            async def run(cmd, cwd):
                """Run a shell command in cwd, return (exit_code, stdout, stderr)."""
                proc = await asyncio.create_subprocess_exec(
                    *cmd,
                    cwd=cwd,
                    stdout=asyncio.subprocess.PIPE,
                    stderr=asyncio.subprocess.PIPE,
                )
                out, err = await proc.communicate()
                return proc.returncode, out.decode().strip(), err.decode().strip()

            results = {}
            name, path = ("integration", integration_dir)
            code, out, err = await run(["git", "reset", "--hard", "HEAD"], cwd=path)
            results[f"{name}_reset"] = {"code": code, "stdout": out, "stderr": err}
            if code != 0:
                logging.error("Reset %s failed: %s", name, err)
                return web.json_response(
                    {
                        "error": f"{name} reset failed",
                        "details": results[name + "_reset"],
                    },
                    status=500,
                )

            code, out, err = await run(["git", "pull"], cwd=path)
            results[f"{name}_pull"] = {"code": code, "stdout": out, "stderr": err}
            if code != 0:
                logging.error("Pull %s failed: %s", name, err)
                return web.json_response(
                    {
                        "error": f"{name} pull failed",
                        "details": results[name + "_pull"],
                    },
                    status=500,
                )

            logging.info("Successfully updated integration")
            return web.json_response(
                {
                    "message": "TIS integrations updated successfully",
                    "results": results,
                }
            )

        except Exception as e:
            logging.error(f"Could Not Update Server: {e}")
            return web.json_response({"error": "Failed to update server"}, status=500)


class BillConfigEndpoint(HomeAssistantView):
    """Save Bill Configurations"""

    url = "/api/bill-config"
    name = "api:bill-config"
    requires_auth = False

    def __init__(self, tis_api: TISApi):
        self.tis_api = tis_api

    async def post(self, request):
        try:
            data = await request.json()

            if not data or "summer_rates" not in data or "winter_rates" not in data:
                logging.error("Required parameters are missing in the request")
                return web.json_response(
                    {"error": "Required parameters are missing"}, status=400
                )

            directory = "/config/custom_components/tis_integration/"
            os.makedirs(directory, exist_ok=True)

            file_name = "bill.json"
            output_file = os.path.join(directory, file_name)

            async with aiofiles.open(output_file, "w") as f:
                await f.write(json.dumps(data, indent=4))

            self.tis_api.bill_configs = data

            # Start reload operations in the background
            asyncio.create_task(self.reload_platforms())

            # Return the response immediately
            return web.json_response({"message": "Bill config saved successfully"})
        except Exception as e:
            logging.error(f"Error saving bill config: {e}")
            return web.json_response(
                {"error": "Failed to save bill config"}, status=500
            )

    async def reload_platforms(self):
        # Reload the platforms
        for entry in self.api.hass.config_entries.async_entries(self.api.domain):
            await self.api.hass.config_entries.async_reload(entry.entry_id)


class GetBillConfigEndpoint(HomeAssistantView):
    """Get Bill Configurations"""

    url = "/api/get-bill-config"
    name = "api:get-bill-config"
    requires_auth = False

    def __init__(self, tis_api: TISApi):
        self.tis_api = tis_api

    async def post(self, request):
        try:
            if self.tis_api.bill_configs:
                configs = self.tis_api.bill_configs
            else:
                configs = await self.tis_api.get_bill_configs()

            logging.info(f"bill configs: {configs}")

            return web.json_response({"config": configs})
        except Exception as e:
            logging.error(f"Error getting bill config: {e}")
            return web.json_response({"error": "Failed to get bill config"}, status=500)


class CMSDataSender:
    """CMS Data class."""

    def __init__(self, external_url: str, hass: HomeAssistant) -> None:
        self.external_url = external_url
        self.hass = hass

    async def send_data(self, data):
        if data is not None:
            try:
                session = async_get_clientsession(self.hass)
                async with session.post(self.external_url, json=data) as response:
                    if response.status != 200:
                        error_text = await response.text()
                        logging.warning(f"Error sending data to CMS: {response.status}")
                        logging.info(f"Error response: {error_text}")
                        return False
                    else:
                        logging.info("Data sent to CMS successfully")
                        return True
            except aiohttp.ClientError as e:
                logging.warning(f"ClientError while sending data to CMS: {e}")
                return False

        return False



from homeassistant.core import HomeAssistant
import logging
import struct

wind_direction_dict = {0x01:"north", 0x02:"north east", 0x04:"east", 0x08:"south east", 0x10:"south", 0x20:"south west", 0x40:"west", 0x80:"north west", }
def big_endian_to_float( value ):
    binary = value.to_bytes(4, 'big')
    float_value = struct.unpack('.>f', binary)
    return float_value

async def handle_weather_feedback(hass: HomeAssistant, info: dict):
    """
    Handle the feedback from a health sensor.
    """
    device_id = info["device_id"]
    wind_direction = wind_direction_dict[int(info["additional_bytes"][3])]
    temperature = big_endian_to_float(int((info["additional_bytes"][4]<<24)|(info["additional_bytes"][5]<<16)|(info["additional_bytes"][6]<<8)|info["additional_bytes"][7]))
    humidity = int(info["additional_bytes"][8])
    wind_speed = big_endian_to_float(int((info["additional_bytes"][9]<<24)|(info["additional_bytes"][10]<<16)|(info["additional_bytes"][11]<<8)|info["additional_bytes"][12]))
    gust_speed = big_endian_to_float(int((info["additional_bytes"][13]<<24)|(info["additional_bytes"][14]<<16)|(info["additional_bytes"][15]<<8)|info["additional_bytes"][16]))
    rainfall = int((info["additional_bytes"][17]<<8)|(info["additional_bytes"][18]))
    lighting = big_endian_to_float(int((info["additional_bytes"][19]<<24)|(info["additional_bytes"][20]<<16)|(info["additional_bytes"][21]<<8)|info["additional_bytes"][22]))
    uv = int(info["additional_bytes"][23])



    event_data = {
        "device_id": device_id,
        "feedback_type": "health_feedback",
        "wind": wind_direction,
        "temperature": temperature,
        "humidity": humidity,
        "wind_speed": wind_speed,
        "gust_speed": gust_speed,
        "rainfall": rainfall,
        "lighting": lighting,
        "uv": uv,
        "additional_bytes": info["additional_bytes"],
    }
    
    try:
        hass.bus.async_fire(str(info["device_id"]), event_data)

    except Exception as e:
        logging.error(f"error in firing event for feedback health: {e}")



from homeassistant.core import HomeAssistant
import struct
import logging


def big_endian_to_num(str_list):
    big_endian = "".join(f"{x:02x}" for x in str_list)
    value = struct.unpack(">f", bytes.fromhex(big_endian))[0]
    return round(value, 1)


async def handle_energy_feedback(hass: HomeAssistant, info: dict):
    """
    Handle the feedback from an energy sensor.
    """
    device_id = info["device_id"]
    channel_num = int(info["additional_bytes"][0]) + 1
    sub_operation = int(info["additional_bytes"][1])

    if sub_operation == 0xDA:
        energy = int(
            (info["additional_bytes"][16] << 8) | (info["additional_bytes"][17])
        )

        event_data = {
            "device_id": device_id,
            "channel_num": channel_num,
            "feedback_type": "monthly_energy_feedback",
            "energy": energy,
            "additional_bytes": info["additional_bytes"],
        }

        try:
            hass.bus.async_fire(str(info["device_id"]), event_data)
        except Exception as e:
            logging.error(f"error in firing event for feedback: {e}")
    elif sub_operation == 0x65:
        try:
            energy = {
                "v1": big_endian_to_num(info["additional_bytes"][3:7]),
                "v2": big_endian_to_num(info["additional_bytes"][7:11]),
                "v3": big_endian_to_num(info["additional_bytes"][11:15]),
                "current_p1": big_endian_to_num(info["additional_bytes"][15:19]),
                "current_p2": big_endian_to_num(info["additional_bytes"][19:23]),
                "current_p3": big_endian_to_num(info["additional_bytes"][23:27]),
                "active_p1": big_endian_to_num(info["additional_bytes"][27:31]),
                "active_p2": big_endian_to_num(info["additional_bytes"][31:35]),
                "active_p3": big_endian_to_num(info["additional_bytes"][35:39]),
                "apparent1": big_endian_to_num(info["additional_bytes"][39:43]),
                "apparent2": big_endian_to_num(info["additional_bytes"][43:47]),
                "apparent3": big_endian_to_num(info["additional_bytes"][47:51]),
                "reactive1": big_endian_to_num(info["additional_bytes"][51:55]),
                "reactive2": big_endian_to_num(info["additional_bytes"][55:59]),
                "reactive3": big_endian_to_num(info["additional_bytes"][59:63]),
                "pf1": big_endian_to_num(info["additional_bytes"][63:67]),
                "pf2": big_endian_to_num(info["additional_bytes"][67:71]),
                "pf3": big_endian_to_num(info["additional_bytes"][71:75]),
                "pa1": big_endian_to_num(info["additional_bytes"][75:79]),
                "pa2": big_endian_to_num(info["additional_bytes"][79:83]),
                "pa3": big_endian_to_num(info["additional_bytes"][83:87]),
                "avg_live_to_neutral": big_endian_to_num(
                    info["additional_bytes"][87:91]
                ),
                "avg_current": big_endian_to_num(info["additional_bytes"][91:95]),
                "sum_current": big_endian_to_num(info["additional_bytes"][95:99]),
                "total_power": big_endian_to_num(info["additional_bytes"][107:111]),
                "total_volt_amps": big_endian_to_num(info["additional_bytes"][115:119]),
                "total_var": big_endian_to_num(info["additional_bytes"][123:127]),
                "total_pf": big_endian_to_num(info["additional_bytes"][127:131]),
                "total_pa": big_endian_to_num(info["additional_bytes"][135:139]),
                "frq": big_endian_to_num(info["additional_bytes"][143:147]),
            }

            event_data = {
                "device_id": device_id,
                "channel_num": channel_num,
                "feedback_type": "energy_feedback",
                "energy": energy,
                "additional_bytes": info["additional_bytes"],
            }

            hass.bus.async_fire(str(info["device_id"]), event_data)
        except Exception as e:
            logging.error(f"error in firing event for feedback: {e}")



from homeassistant.core import HomeAssistant
import logging


async def handle_floor_binary_feedback(hass: HomeAssistant, info: dict):
    
    
    #check sub_operation or number
    number = info["additional_bytes"][0]
    state = info["additional_bytes"][3]
    temp = info["additional_bytes"][5]

    event_data = {
    "device_id": info["device_id"],
    "feedback_type":"floor_update",
    "heater_number": number,
    "state": state,
    "temp": temp,
    }
            

    try:
        hass.bus.async_fire(str(info["device_id"]), event_data)
        
    except Exception as e:
        logging.error(f"error in firing event: {e}")



"""Class for handling the UDP protocol"""

from ...BytesHelper import build_packet
from typing import List, Literal, Tuple


class TISPacket:
    """
    Class representing a Packet.

    :param device_id: List of integers representing the device ID.
    :param operation_code: List of integers representing the operation code.
    :param source_ip: Source IP address as a string.
    :param destination_ip: Destination IP address as a string.
    :param additional_bytes: Optional list of additional bytes.
    """

    def __init__(
        self,
        device_id: List[int],
        operation_code: List[int],
        source_ip: str,
        destination_ip: str,
        additional_bytes: List[int] = None,
    ):
        if additional_bytes is None:
            additional_bytes = []
        self.device_id = device_id
        self.operation_code = operation_code
        self.source_ip = source_ip
        self.destination_ip = destination_ip
        self.additional_bytes = additional_bytes
        self._packet = build_packet(
            ip_address=self.source_ip,
            device_id=self.device_id,
            operation_code=self.operation_code,
            additional_packets=self.additional_bytes,
        )

    def __str__(self) -> str:
        return f"Packet: {self._packet}"

    def __repr__(self) -> str:
        return f"Packet: {self._packet}"

    def __bytes__(self) -> bytes:
        return bytes(self._packet)


class TISProtocolHandler:
    OPERATION_CONTROL = [0x00, 0x31]
    OPERATION_CONTROL_UPDATE = [0x00, 0x33]
    OPERATION_GET_TEMP = [0xE3, 0xE7]
    OPERATION_GET_HEALTH = [0x20, 0x24]
    OPERATION_DISCOVERY = [0x00, 0x0E]
    OPERATION_CONTROL_SECURITY = [0x01, 0x04]
    OPERATION_CONTROL_AC = [0xE0, 0xEE]
    OPERATION_AC_UPDATE = [0xE0, 0xEC]
    OPERATION_FLOOR_UPDATE = [0x19, 0x44]
    OPERATION_FLOOR_CONTROL = [0xE3, 0xD8]
    OPERATION_GET_WEATHER = [0x20, 0x20]
    OPERATION_SECURITY_UPDATE = [0x01, 0x1E]
    OPERATION_ANALOG_UPDATE = [0xEF, 0x00]
    OPERATION_ENERGY_UPDATE = [0x20, 0x10]
    OPERATION_UNIVERSAL_SWITCH = [0xE0, 0x1C]

    def __init__(self) -> None:
        """Initialize a ProtocolHandler instance."""
        pass

    def generate_control_on_packet(self, entity) -> TISPacket:
        """
        Generate a packet to switch on the device.

        :param entity: The entity object containing device information.
        :return: A Packet instance.
        """
        return TISPacket(
            device_id=entity.device_id,
            operation_code=TISProtocolHandler.OPERATION_CONTROL,
            source_ip=entity.api.host,
            destination_ip=entity.gateway,
            additional_bytes=[entity.channel_number, 0x64, 0x00, 0x00],
        )

    def generate_control_off_packet(self, entity) -> TISPacket:
        """
        Generate a packet to switch off the device.

        :param entity: The entity object containing device information.
        :return: A Packet instance.
        """
        return TISPacket(
            device_id=entity.device_id,
            operation_code=TISProtocolHandler.OPERATION_CONTROL,
            source_ip=entity.api.host,
            destination_ip=entity.gateway,
            additional_bytes=[entity.channel_number, 0x00, 0x00, 0x00],
        )

    def generate_control_update_packet(self, entity) -> TISPacket:
        """
        Generate a packet to update the device control.

        :param entity: The entity object containing device information.
        :return: A Packet instance.
        """
        return TISPacket(
            device_id=entity.device_id,
            operation_code=TISProtocolHandler.OPERATION_CONTROL_UPDATE,
            source_ip=entity.api.host,
            destination_ip=entity.gateway,
            additional_bytes=[],
        )

    def generate_temp_sensor_update_packet(self, entity) -> TISPacket:
        """
        Generate a packet to update the temperature sensor.

        :param entity: The entity object containing device information.
        :return: A Packet instance.
        """
        return TISPacket(
            device_id=entity.device_id,
            operation_code=TISProtocolHandler.OPERATION_GET_TEMP,
            source_ip=entity.api.host,
            destination_ip=entity.gateway,
            additional_bytes=[0x00],
        )

    def generate_health_sensor_update_packet(self, entity) -> TISPacket:
        """
        Generate a packet to update the health sensor.

        :param entity: The entity object containing device information.
        :return: A Packet instance.
        """
        return TISPacket(
            device_id=entity.device_id,
            operation_code=TISProtocolHandler.OPERATION_GET_HEALTH,
            source_ip=entity.api.host,
            destination_ip=entity.gateway,
            additional_bytes=[0x14, 0x00],
        )

    def generate_weather_sensor_update_packet(self, entity) -> TISPacket:
        """
        Generate a packet to update the weather sensor.

        :param entity: The entity object containing device information.
        :return: A Packet instance.
        """
        return TISPacket(
            device_id=entity.device_id,
            operation_code=TISProtocolHandler.OPERATION_GET_WEATHER,
            source_ip=entity.api.host,
            destination_ip=entity.gateway,
            additional_bytes=[0x4, 0x00],
        )

    def generate_discovery_packet(self) -> TISPacket:
        """
        Generate a packet to discover devices on the network.

        :return: A Packet instance.
        """
        return TISPacket(
            device_id=[0xFF, 0xFF],
            operation_code=TISProtocolHandler.OPERATION_DISCOVERY,
            source_ip="0.0.0.0",
            destination_ip="0.0.0.0",
            additional_bytes=[],
        )

    def generate_light_control_packet(self, entity, brightness: int) -> TISPacket:
        """
        Generate packets to control a light.
        :param entity: The entity object containing device information.
        :param brightness: An integer representing the brightness level.
        :return: A Packet instance.
        """
        return TISPacket(
            device_id=entity.device_id,
            operation_code=TISProtocolHandler.OPERATION_CONTROL,
            source_ip=entity.api.host,
            destination_ip=entity.gateway,
            additional_bytes=[entity.channel_number, brightness, 0x00, 0x00],
        )

    def generate_rgb_light_control_packet(
        self, entity, color: Tuple[int, int, int]
    ) -> Tuple[TISPacket]:
        """
        Generate packets to control an RGB light.
        :param entity: The entity object containing device information.
        :param color: A tuple of integers representing the RGB color.
        :return: A tuple of Packet instances.
        """
        return (
            TISPacket(
                device_id=entity.device_id,
                operation_code=TISProtocolHandler.OPERATION_CONTROL,
                source_ip=entity.api.host,
                destination_ip=entity.gateway,
                additional_bytes=[entity.r_channel, color[0], 0x00, 0x00],
            ),
            TISPacket(
                device_id=entity.device_id,
                operation_code=TISProtocolHandler.OPERATION_CONTROL,
                source_ip=entity.api.host,
                destination_ip=entity.gateway,
                additional_bytes=[entity.g_channel, color[1], 0x00, 0x00],
            ),
            TISPacket(
                device_id=entity.device_id,
                operation_code=TISProtocolHandler.OPERATION_CONTROL,
                source_ip=entity.api.host,
                destination_ip=entity.gateway,
                additional_bytes=[entity.b_channel, color[2], 0x00, 0x00],
            ),
        )

    def generate_rgbw_light_control_packet(
        self, entity, color: Tuple[int, int, int, int]
    ) -> Tuple[TISPacket]:
        """
        Generate packets to control an RGBW light.
        :param entity: The entity object containing device information.
        :param color: A tuple of integers representing the RGBW color.
        :return: A tuple of Packet instances.
        """
        return (
            TISPacket(
                device_id=entity.device_id,
                operation_code=TISProtocolHandler.OPERATION_CONTROL,
                source_ip=entity.api.host,
                destination_ip=entity.gateway,
                additional_bytes=[entity.r_channel, color[0], 0x00, 0x00],
            ),
            TISPacket(
                device_id=entity.device_id,
                operation_code=TISProtocolHandler.OPERATION_CONTROL,
                source_ip=entity.api.host,
                destination_ip=entity.gateway,
                additional_bytes=[entity.g_channel, color[1], 0x00, 0x00],
            ),
            TISPacket(
                device_id=entity.device_id,
                operation_code=TISProtocolHandler.OPERATION_CONTROL,
                source_ip=entity.api.host,
                destination_ip=entity.gateway,
                additional_bytes=[entity.b_channel, color[2], 0x00, 0x00],
            ),
            TISPacket(
                device_id=entity.device_id,
                operation_code=TISProtocolHandler.OPERATION_CONTROL,
                source_ip=entity.api.host,
                destination_ip=entity.gateway,
                additional_bytes=[entity.w_channel, color[3], 0x00, 0x00],
            ),
        )

    def generate_no_pos_cover_packet(
        self, entity, mode: Literal["open", "close", "stop"]
    ) -> tuple[TISPacket, TISPacket]:
        if mode == "open":
            return (
                TISPacket(
                    device_id=entity.device_id,
                    operation_code=TISProtocolHandler.OPERATION_CONTROL,
                    source_ip=entity.api.host,
                    destination_ip=entity.gateway,
                    additional_bytes=[entity.up_channel_number, 0x64, 0x00, 0x00],
                ),
                TISPacket(
                    device_id=entity.device_id,
                    operation_code=TISProtocolHandler.OPERATION_CONTROL,
                    source_ip=entity.api.host,
                    destination_ip=entity.gateway,
                    additional_bytes=[entity.down_channel_number, 0x00, 0x00, 0x00],
                ),
            )
        elif mode == "close":
            return (
                TISPacket(
                    device_id=entity.device_id,
                    operation_code=TISProtocolHandler.OPERATION_CONTROL,
                    source_ip=entity.api.host,
                    destination_ip=entity.gateway,
                    additional_bytes=[entity.up_channel_number, 0x00, 0x00, 0x00],
                ),
                TISPacket(
                    device_id=entity.device_id,
                    operation_code=TISProtocolHandler.OPERATION_CONTROL,
                    source_ip=entity.api.host,
                    destination_ip=entity.gateway,
                    additional_bytes=[entity.down_channel_number, 0x64, 0x00, 0x00],
                ),
            )

        elif mode == "stop":
            return (
                TISPacket(
                    device_id=entity.device_id,
                    operation_code=TISProtocolHandler.OPERATION_CONTROL,
                    source_ip=entity.api.host,
                    destination_ip=entity.gateway,
                    additional_bytes=[entity.up_channel_number, 0x00, 0x00, 0x00],
                ),
                TISPacket(
                    device_id=entity.device_id,
                    operation_code=TISProtocolHandler.OPERATION_CONTROL,
                    source_ip=entity.api.host,
                    destination_ip=entity.gateway,
                    additional_bytes=[entity.down_channel_number, 0x00, 0x00, 0x00],
                ),
            )

    def generate_control_security_packet(self, entity, mode) -> TISPacket:
        """
        Generate a packet to set the security mode.

        vacation=1
        Away=2
        Night=3
        Disarm=6
        """
        return TISPacket(
            device_id=entity.device_id,
            operation_code=TISProtocolHandler.OPERATION_CONTROL_SECURITY,
            source_ip=entity.api.host,
            destination_ip=entity.gateway,
            additional_bytes=[entity.channel_number, mode],
        )

    def generate_update_security_packet(self, entity) -> TISPacket:
        """
        Generate a packet to set request security update.
        """
        return TISPacket(
            device_id=entity.device_id,
            operation_code=TISProtocolHandler.OPERATION_SECURITY_UPDATE,
            source_ip=entity.api.host,
            destination_ip=entity.gateway,
            additional_bytes=[entity.channel_number],
        )

    def generate_ac_control_packet(
        self,
        entity,
        temperature_ranges: dict,
        fan_modes: dict,
        target_state: str | None = None,
        target_temperature: float | None = None,
        target_mode: str | None = None,  # noqa: F821 # type: ignore
        target_fan_mode: str | None = None,
    ) -> TISPacket:
        # Determine the target values, falling back to class attributes if not provided
        if not target_state:
            if entity._attr_state == "unknown":
                target_state = False
            else:
                target_state = entity._attr_state

        target_temperature = (
            target_temperature
            if target_temperature is not None
            else entity._attr_target_temperature
        )
        target_mode = target_mode if target_mode is not None else entity.hvac_mode
        target_fan_mode = (
            target_fan_mode if target_fan_mode is not None else entity._attr_fan_mode
        )
        # Convert target temperature to byte
        target_temperature_byte = (
            int(target_temperature) if target_temperature else 0x00
        )
        # Construct the additional bytes for the packet
        additional_bytes = [
            entity.ac_number,
            int(target_state == "on"),
            target_temperature_byte,
            int(
                (temperature_ranges[target_mode]["packet_mode_index"] << 4)
                | fan_modes[target_fan_mode]
            ),
            0x01,
            target_temperature_byte,
            target_temperature_byte,
            target_temperature_byte,
            0x00,
        ]

        # Generate and return the packet
        return TISPacket(
            device_id=entity.device_id,
            operation_code=self.OPERATION_CONTROL_AC,
            source_ip=entity.api.host,
            destination_ip=entity.gateway,
            additional_bytes=additional_bytes,
        )

    def generate_ac_update_packet(self, entity) -> TISPacket:
        return TISPacket(
            device_id=entity.device_id,
            operation_code=self.OPERATION_AC_UPDATE,
            source_ip=entity.api.host,
            destination_ip=entity.gateway,
            additional_bytes=[entity.ac_number],
        )

    def generate_floor_update_packet(self, entity) -> TISPacket:
        return TISPacket(
            device_id=entity.device_id,
            operation_code=self.OPERATION_FLOOR_UPDATE,
            source_ip=entity.api.host,
            destination_ip=entity.gateway,
            additional_bytes=[entity.heater_number],
        )

    def generate_floor_on_off_packet(self, entity, state: int) -> TISPacket:
        return TISPacket(
            device_id=entity.device_id,
            operation_code=self.OPERATION_FLOOR_CONTROL,
            source_ip=entity.api.host,
            destination_ip=entity.gateway,
            additional_bytes=[(entity.heater_number + 0x22), 0x14, state],
        )

    def generate_floor_set_temp_packet(
        self, entity, target_temperature: int
    ) -> TISPacket:
        return TISPacket(
            device_id=entity.device_id,
            operation_code=self.OPERATION_FLOOR_CONTROL,
            source_ip=entity.api.host,
            destination_ip=entity.gateway,
            additional_bytes=[(entity.heater_number + 0x22), 0x18, target_temperature],
        )

    def generate_update_analog_packet(self, entity) -> TISPacket:
        """
        Generate a packet to set request analog update.
        """
        return TISPacket(
            device_id=entity.device_id,
            operation_code=TISProtocolHandler.OPERATION_ANALOG_UPDATE,
            source_ip=entity.api.host,
            destination_ip=entity.gateway,
            additional_bytes=[],
        )

    def generate_update_monthly_energy_packet(self, entity) -> TISPacket:
        """
        Generate a packet to set request energy meter update.
        """
        return TISPacket(
            device_id=entity.device_id,
            operation_code=TISProtocolHandler.OPERATION_ENERGY_UPDATE,
            source_ip=entity.api.host,
            destination_ip=entity.gateway,
            additional_bytes=[entity.channel_number - 1, 0xDA, 0x64],
        )

    def generate_update_energy_packet(self, entity) -> TISPacket:
        """
        Generate a packet to set request energy meter update.
        """
        return TISPacket(
            device_id=entity.device_id,
            operation_code=TISProtocolHandler.OPERATION_ENERGY_UPDATE,
            source_ip=entity.api.host,
            destination_ip=entity.gateway,
            additional_bytes=[entity.channel_number - 1, 0x65],
        )

    def generate_universal_switch_packet(self, entity) -> TISPacket:
        """
        Generate a packet to set request energy meter update.
        """
        return TISPacket(
            device_id=entity.device_id,
            operation_code=TISProtocolHandler.OPERATION_UNIVERSAL_SWITCH,
            source_ip=entity.api.host,
            destination_ip=entity.gateway,
            additional_bytes=[entity.channel_number, entity.universal_type],
        )



from homeassistant.core import HomeAssistant
import logging

AC_NUMBER_MAP = {0x19: 0, 0x1A: 1, 0x1B: 2, 0x1C: 3, 0x1D: 4, 0x1E: 5, 0x1F: 6, 0x20: 7}
FLOOR_NUMBER_MAP = {0x22: 0, 0x23: 1, 0x24: 2, 0x25: 3}

async def handle_climate_binary_feedback(hass: HomeAssistant, info: dict):
    
    #NOTE: Sometimes the packet contains number of floorheater and sometimes no floorheater num is given
    #check sub_operation or number
    if info["additional_bytes"][0] <= 0x18:
        sub_operation = info["additional_bytes"][0]
        operation_value = info["additional_bytes"][1]

        if info["additional_bytes"][0] < 0x14:
            
            feedback_type = "ac_feedback"
            number = 0
            
        else:
            feedback_type = "floor_feedback"
            number = 0
                

    else:
        ac_number = AC_NUMBER_MAP.get(info["additional_bytes"][0], None)
        floor_number = FLOOR_NUMBER_MAP.get(info["additional_bytes"][0], None)
        sub_operation = info["additional_bytes"][1]
        operation_value = info["additional_bytes"][2]

        if ac_number is not None:
            feedback_type = "ac_feedback"
            number = ac_number

        else:
            feedback_type = "floor_feedback"
            number = floor_number        
        
    event_data = {
    "device_id": info["device_id"],
    "feedback_type": feedback_type,
    "number": number,
    "sub_operation": sub_operation,
    "operation_value": operation_value,
    }
            

    try:
        hass.bus.async_fire(str(info["device_id"]), event_data)
    
    except Exception as e:
        logging.error(f"error in firing event: {e}")



from homeassistant.core import HomeAssistant  # type: ignore
import logging
import asyncio
from TISControlProtocol.shared import ack_events


async def handle_security_update_feedback (hass: HomeAssistant, info: dict):
    # remove auxilary bytes which represents number of scenarios
    channel_number = info["additional_bytes"][0]
    mode = info["additional_bytes"][1]
    event_data = {
        "device_id": info["device_id"],
        "feedback_type": "security_update",
        "additional_bytes": info["additional_bytes"],
        "channel_number": channel_number,
        "mode": mode,
    }
    try:
        hass.bus.async_fire(str(info["device_id"]), event_data)
    except Exception as e:
        logging.error(f"error in firing event: {e}")

    try:
        event: asyncio.Event = ack_events.get(
            (
                tuple(info["device_id"]),
                (0x01, 0x1E),
                int(channel_number),
            )
        )
        if event is not None:
            event.set()
    except Exception as e:
        logging.error(f'error in setting event for: {info["device_id"]}, {e}')


from homeassistant.core import HomeAssistant
import logging


async def handle_health_feedback(hass: HomeAssistant, info: dict):
    """
    Handle the feedback from a health sensor.
    """

    device_id = info["device_id"]

    # this try to know if this device is 10F or health sensor, for 10F it will make all readings 0 except LUX
    # for eco2,tvoc,co states >>  0: Not Ready, 1: Excellent, 2: Normal, 3: Low Risk, 4: Med Risk, 5: High Rish

    try:
        lux = int((info["additional_bytes"][5] << 8) | (info["additional_bytes"][6]))
        noise = int((info["additional_bytes"][7] << 8) | (info["additional_bytes"][8]))
        eco2 = int((info["additional_bytes"][9] << 8) | (info["additional_bytes"][10]))
        tvoc = int((info["additional_bytes"][11] << 8) | (info["additional_bytes"][12]))
        temp = int(info["additional_bytes"][13])
        humidity = int(info["additional_bytes"][14])
        co = int((info["additional_bytes"][27] << 8) | (info["additional_bytes"][28]))
        # There are 2 additional bytes for flags not mentioned in the docs.
        eco2_state = int(info["additional_bytes"][31])
        tvoc_state = int(info["additional_bytes"][32])
        co_state = int(info["additional_bytes"][33])

    except Exception as _:
        tvoc = 0
        noise = 0
        temp = 0
        humidity = 0
        eco2 = 0
        co = 0
        eco2_state = 0
        tvoc_state = 0
        co_state = 0

    event_data = {
        "device_id": device_id,
        "feedback_type": "health_feedback",
        "lux": lux,
        "noise": noise,
        "eco2": eco2,
        "tvoc": tvoc,
        "co": co,
        "temp": temp,
        "humidity": humidity,
        "eco2_state": eco2_state,
        "tvoc_state": tvoc_state,
        "co_state": co_state,
        "additional_bytes": info["additional_bytes"],
    }

    try:
        hass.bus.async_fire(str(info["device_id"]), event_data)
    except Exception as e:
        logging.error(f"error in firing event for feedback health: {e}")



from homeassistant.core import HomeAssistant
import logging

#TODO get a way to set 4 sensors together
async def handle_analog_feedback(hass: HomeAssistant, info: dict):
    """
    Handle the feedback from an analog sensor.
    """
    device_id = info["device_id"]
    channels_num = int(info["additional_bytes"][0])
    analog = info["additional_bytes"][1:channels_num + 1]

    event_data = {
        "device_id": device_id,
        "feedback_type": "analog_feedback",
        "analog": analog,
        "additional_bytes": info["additional_bytes"],
    }

    try:
        hass.bus.async_fire(str(info["device_id"]), event_data)
    except Exception as e:
        logging.error(f"error in firing event for feedback: {e}")



from homeassistant.core import HomeAssistant
import logging

import asyncio
from TISControlProtocol.shared import ack_events



async def handle_security_feedback(hass: HomeAssistant, info: dict):
    """
    Handle the feedback from a security sensor.
    """
    channel_number = info["additional_bytes"][0]
    mode = info["additional_bytes"][1]
    event_data = {
        "device_id": info["device_id"],
        "feedback_type": "security_feedback",
        "additional_bytes": info["additional_bytes"],
        "channel_number": channel_number,
        "mode": mode,
    }
    try:
        hass.bus.async_fire(str(info["device_id"]), event_data)
        logging.info(
            f"control response event fired for {info['device_id']}, additional bytes: {info['additional_bytes']}"
        )
    except Exception as e:
        logging.error(f"error in firing event for feedback security: {e}")

    try:
        event: asyncio.Event = ack_events.get(
            (
                tuple(info["device_id"]),
                (0x01, 0x04),
                int(channel_number),
            )
        )
        if event is not None:
            event.set()
    except Exception as e:
        logging.error(f"error in setting event for {info["device_id"]}: {e}")



from homeassistant.core import HomeAssistant
import logging


async def handle_luna_temp_feedback(hass: HomeAssistant, info: dict):
    """
    Handle the feedback from a Luna temperature sensor.
    """
    device_id = info["device_id"]
    temperature = int(info["additional_bytes"][1])

    event_data = {
        "device_id": device_id,
        "feedback_type": "temp_feedback",
        "temp": temperature,
        "additional_bytes": info["additional_bytes"],
    }

    try:
        hass.bus.async_fire(str(info["device_id"]), event_data)
    except Exception as e:
        logging.error(f"error in firing event for feedback: {e}")
